# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dynamic Model Parser."""

import importlib
import warnings
from typing import List

from pydantic import Field, create_model

from trestle.core import const
from trestle.core.base_model import OscalBaseModel
from trestle.core.err import TrestleError
from trestle.utils import fs
from trestle.utils import log

logger = log.get_logger()


def parse_dict(data: dict, model_name: str):
    """Load a model from the data dict.

    Argument:
        model_name: it should be of the form <module>.<class>
                    <class> should be a Pydantic class that supports `parse_obj` method
    """
    warnings.warn('trestle.parser functions are deprecated', DeprecationWarning)
    if data is None:
        raise TrestleError('data name is required')

    if model_name is None:
        raise TrestleError('model_name is required')

    parts = model_name.split('.')
    class_name = parts.pop()
    module_name = '.'.join(parts)

    logger.debug(f'Loading class "{class_name}" from "{module_name}"')
    module = importlib.import_module(module_name)
    mclass = getattr(module, class_name)
    if mclass is None:
        raise TrestleError(f'class "{class_name}" could not be found in "{module_name}"')

    instance = mclass.parse_obj(data)
    return instance


def root_key(data: dict):
    """Find root model name in the data."""
    warnings.warn('trestle.parser functions are deprecated', DeprecationWarning)
    if len(data.items()) == 1:
        return next(iter(data))

    raise TrestleError('data does not contain a root key')


def to_class_name(name: str) -> str:
    """Convert to pascal class name."""
    warnings.warn('trestle.parser functions are deprecated', DeprecationWarning)
    if name.find('-') != -1:
        parts = name.split('-')

        for i, part in enumerate(parts):
            parts[i] = part.capitalize()

        name = ''.join(parts)
        return name

    chars = list(name)
    chars[0] = chars[0].upper()
    return ''.join(chars)


def to_full_model_name(root_key: str, name: str = None):
    """Find model name from the root_key in the file."""
    warnings.warn('trestle.parser functions are deprecated', DeprecationWarning)
    try:
        # process root key and extract model name
        module_name = root_key.lower()
        if root_key.find('-') != -1:
            parts = root_key.split('-')
            module_name = parts[0]

            for i, part in enumerate(parts):
                parts[i] = part.capitalize()

            name = ''.join(parts)

        # check for module with the root-key
        module = importlib.import_module(f'{const.PACKAGE_OSCAL}.{module_name}')

        # prepare class name
        if name is None:
            name = module_name
        class_name = to_class_name(name)

        # check if class exists in the module or not
        if getattr(module, class_name) is not None:
            return f'{const.PACKAGE_OSCAL}.{module_name}.{class_name}'
    except ModuleNotFoundError as ex:
        logger.error(f'Module {module_name} not found: {ex}')
        pass

    return None


def parse_file(file_name: str, model_name: str):
    """Load a model from the file.

    Argument:
        model_name: it should be of the form <module>.<class>
                    <class> should be a Pydantic class that supports `parse_obj` method
    """
    warnings.warn('trestle.parser functions are deprecated', DeprecationWarning)
    if file_name is None:
        raise TrestleError('file_name is required')

    data = fs.load_file(file_name)
    rkey = root_key(data)
    if model_name is None:
        model_name = to_full_model_name(rkey)
    return parse_dict(data[rkey], model_name)


def wrap_for_output(model: OscalBaseModel) -> OscalBaseModel:
    """Dynamically wrap for output a class such that the correct string is provided."""
    warnings.warn(
        'trestle.parser functions are deprecated. wrap_for_output built into OSCALBaseModel.', DeprecationWarning
    )
    # TODO: Refactor based on next method.
    class_name = model.__class__.__name__
    # It would be nice to pass through the description but I can't seem to and
    # it does not affect the output
    dynamic_passer = {}
    dynamic_passer[class_to_oscal(class_name, 'field')] = (
        model.__class__,
        Field(model, title=class_to_oscal(class_name, 'field'), alias=class_to_oscal(class_name, 'json'))
    )
    wrapper_model = create_model(class_name, __base__=OscalBaseModel, **dynamic_passer)
    # Default behaviour is strange here.
    wrapped_model = wrapper_model(**{class_to_oscal(class_name, 'json'): model})
    return wrapped_model


def wrap_for_input(raw_class):
    """In this instance we are wrapping an actual OSCAL class not an instance."""
    warnings.warn(
        'trestle.parser functions are deprecated. wrap_for_input built ito OSCALBaseModel', DeprecationWarning
    )
    # TODO: Check behaviour is fine when no
    class_name = raw_class.__name__
    dynamic_passer = {}
    dynamic_passer[class_to_oscal(
        class_name, 'field'
    )] = (raw_class, Field(..., title=class_to_oscal(class_name, 'json'), alias=class_to_oscal(class_name, 'json')))
    wrapper_model = create_model('Wrapped' + class_name, __base__=OscalBaseModel, **dynamic_passer)
    return wrapper_model


def class_to_oscal(class_name: str, mode: str) -> str:
    """
    Return oscal json or field element name based on class name.

    This is applicable when asking for a singular element.
    """
    warnings.warn('trestle.parser functions are deprecated. class_to_oscal now contained in utils', DeprecationWarning)
    parts = pascal_case_split(class_name)
    if mode == 'json':
        return '-'.join(map(str.lower, parts))
    elif mode == 'field':
        return '_'.join(map(str.lower, parts))
    else:
        raise TrestleError('Bad option')


def pascal_case_split(pascal_str: str) -> List[str]:
    """Parse a pascal case string (e.g. a ClassName) and return a list of strings."""
    warnings.warn(
        'trestle.parser functions are deprecated. pascal_case_split now contained in utils', DeprecationWarning
    )
    start_idx = [i for i, e in enumerate(pascal_str) if e.isupper()] + [len(pascal_str)]
    return [pascal_str[x:y] for x, y in zip(start_idx, start_idx[1:])]



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common logging utilities."""
import logging

# Singleton logger instance
# All other CLI sub module will inherit settings of this logger as long as
# sub-module instantiates a logger with a prefix 'trestle' or __name__
_logger = logging.getLogger('trestle')


def init(level=logging.DEBUG) -> logging.Logger:
    """Initialize the logger."""
    # create logger
    _logger.setLevel(level)

    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(level)

    # create formatter
    formatter = logging.Formatter('%(asctime)s %(name)s:%(lineno)d %(levelname)s: %(message)s')

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    _logger.addHandler(ch)

    return _logger


def set_level(level=logging.DEBUG):
    """Set log level."""
    _logger.setLevel(level)
    return _logger


def get_logger(level=logging.DEBUG):
    """Get the trestle default logger."""
    return set_level(level)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle cmd utils module."""
from typing import List

from trestle.core.commands import cmd_utils
from trestle.core.models.elements import ElementPath


def prepare_expected_element_paths(element_args: List[str]) -> List[ElementPath]:
    """Prepare a list of ElementPath from a list of path strings."""
    element_paths: List[ElementPath] = []
    for element_arg in element_args:
        element_paths.append(ElementPath(element_arg))

    return element_paths


def test_parse_element_arg():
    """Unit test parse a single element arg."""
    element_arg = 'target-definition.targets'
    expected_paths: List[ElementPath] = prepare_expected_element_paths(['target-definition.targets'])
    element_paths: List[ElementPath] = cmd_utils.parse_element_arg(element_arg)
    assert expected_paths == element_paths

    element_arg = 'target-definition.targets.*'
    expected_paths: List[ElementPath] = prepare_expected_element_paths(['target-definition.targets.*'])
    element_paths: List[ElementPath] = cmd_utils.parse_element_arg(element_arg)
    assert expected_paths == element_paths

    element_arg = 'catalog.groups.*.controls.*.controls.*'
    p1 = ElementPath('catalog.groups.*')
    p2 = ElementPath('controls.*', parent_path=p1)
    p3 = ElementPath('controls.*', parent_path=p2)
    expected_paths: List[ElementPath] = [p1, p2, p3]
    element_paths: List[ElementPath] = cmd_utils.parse_element_arg(element_arg)
    assert expected_paths == element_paths


def test_parse_element_args():
    """Unit test parse multiple element args."""
    element_args = ['catalog.metadata', 'catalog.groups', 'catalog.controls']
    p0 = ElementPath('catalog.metadata')
    p1 = ElementPath('catalog.groups')
    p2 = ElementPath('catalog.controls')
    expected_paths: List[ElementPath] = [p0, p1, p2]
    element_paths = cmd_utils.parse_element_args(element_args)
    assert expected_paths == element_paths

    # element args with wildcard
    element_args = ['catalog.metadata', 'catalog.groups.*.controls.*.controls.*', 'catalog.controls.*.controls.*']
    p0 = ElementPath('catalog.metadata')
    p1 = ElementPath('catalog.groups.*')
    p2 = ElementPath('controls.*', parent_path=p1)
    p3 = ElementPath('controls.*', parent_path=p2)
    p4 = ElementPath('catalog.controls.*')
    p5 = ElementPath('controls.*', parent_path=p4)
    expected: List[ElementPath] = [p0, p1, p2, p3, p4, p5]
    assert cmd_utils.parse_element_args(element_args) == expected



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle command related utilities."""
from typing import List

from trestle.core.base_model import OscalBaseModel
from trestle.core.err import TrestleError
from trestle.core.models.elements import ElementPath


def get_model(file_path: str) -> OscalBaseModel:
    """Get the model specified by the file."""
    raise NotImplementedError()


def parse_element_args(element_args: List[str]) -> List[ElementPath]:
    """Parse element args into a list of ElementPath."""
    if not isinstance(element_args, list):
        raise TrestleError(f'Input element_paths must be a list, but found {element_args.__class__}')

    element_paths: List[ElementPath] = []
    for element_arg in element_args:
        paths = parse_element_arg(element_arg)
        element_paths.extend(paths)

    return element_paths


def parse_element_arg(element_arg: str) -> List[ElementPath]:
    """Parse an element arg stirng into a list of ElementPath."""
    element_paths: List[ElementPath] = []

    # search for wildcards and create paths with its parent path
    last_pos: int = -1
    prev_element_path = None
    for cur_pos, c in enumerate(element_arg):
        if c == ElementPath.WILDCARD:
            # extract the path string including wildcard
            start = last_pos + 1
            end = cur_pos + 1
            p = element_arg[start:end]

            # create and append elment_path
            element_path = ElementPath(p, parent_path=prev_element_path)
            element_paths.append(element_path)

            # store values for next cycle
            prev_element_path = element_path
            last_pos = end

    # if there was no wildcard in the path, it is just a single path
    # so create the path
    if last_pos == -1:
        element_path = ElementPath(element_arg)
        element_paths.append(element_path)

    return element_paths



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for json and yaml manipulations of oscal files."""

import datetime
import pathlib

import pytest

import trestle.oscal.target as ostarget

import yaml

yaml_path = pathlib.Path('tests/data/yaml/')
json_path = pathlib.Path('tests/data/json/')
encoding = 'utf8'
loader = yaml.Loader


def test_yaml_load():
    """Test yaml load."""
    # happy path
    read_file = (yaml_path / 'good_simple.yaml').open('r', encoding=encoding)
    obj = yaml.load(read_file, Loader=loader)
    assert obj is not None

    # unhappy path
    with pytest.raises(yaml.parser.ParserError):
        read_file = (yaml_path / 'bad_simple.yaml').open('r', encoding=encoding)
        obj = yaml.load(read_file, Loader=loader)


def test_yaml_dump(tmpdir):
    """Test yaml load and dump."""
    target_name = 'good_target.yaml'
    tmp_path = pathlib.Path(tmpdir)

    # happy path
    read_file = (yaml_path / target_name).open('r', encoding=encoding)
    target = yaml.load(read_file, Loader=loader)
    assert target is not None

    dump_name = tmp_path / target_name
    write_file = dump_name.open('w', encoding=encoding)
    yaml.dump(target, write_file)
    read_file = dump_name.open('r', encoding=encoding)
    saved_target = yaml.load(read_file, Loader=loader)
    assert saved_target is not None

    assert saved_target == target


def test_oscal_model(tmpdir):
    """Test pydantic oscal model."""
    good_target_name = 'good_target.yaml'
    tmpdir = pathlib.Path(tmpdir)

    tmp_path = pathlib.Path(tmpdir)

    # load good target
    read_file = yaml_path / good_target_name
    assert read_file.exists()
    target = ostarget.TargetDefinition.oscal_read(read_file)
    assert target is not None

    # write the oscal target def out as yaml
    dump_name = tmp_path / good_target_name
    target.oscal_write(dump_name)

    # read it back in
    target_reload = ostarget.TargetDefinition.oscal_read(dump_name)
    assert target_reload is not None

    # confirm same
    assert target == target_reload

    # confirm it really is checking the time
    target_reload.metadata.last_modified = datetime.datetime.now()
    assert target != target_reload

    # load good target with different timezone
    read_file = yaml_path / 'good_target_diff_tz.yaml'
    target_diff_tz = ostarget.TargetDefinition.oscal_read(read_file)
    assert target_diff_tz is not None

    # confirm same since different timezones but same utc time
    assert target == target_diff_tz

    # try to load file with no timezone specified
    read_file = yaml_path / 'bad_target_no_tz.yaml'

    # confirm the load fails because it is invalid without timezone specified
    try:
        _ = ostarget.TargetDefinition.oscal_read(read_file)
    except Exception:
        assert True
    else:
        assert AssertionError()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Split Command."""
from typing import List

from ilcli import Command

from trestle.core import const
from trestle.core.base_model import OscalBaseModel
from trestle.core.err import TrestleError
from trestle.core.models.elements import ElementPath
from trestle.core.models.plans import Plan

from . import cmd_utils


class SplitCmd(Command):
    """Split subcomponents on a trestle model."""

    name = 'split'

    def _init_arguments(self):
        self.add_argument(
            f'-{const.ARG_FILE_SHORT}',
            f'--{const.ARG_FILE}',
            help=const.ARG_DESC_FILE + ' to split.',
        )
        self.add_argument(
            f'-{const.ARG_ELEMENT_SHORT}',
            f'--{const.ARG_ELEMENT}',
            help=const.ARG_DESC_ELEMENT + ' to split.',
        )

    def _run(self, args):
        """Split an OSCAL file into elements."""
        # get the Model
        if args[const.ARG_FILE] is None:
            raise TrestleError(f'Argument "-{const.ARG_FILE_SHORT}" is required')

        model: OscalBaseModel = cmd_utils.get_model(args[const.ARG_FILE])
        element_paths: List[ElementPath] = cmd_utils.parse_element_args(args[const.ARG_ELEMENT])

        split_plan = self._split_model(model, element_paths)

        try:
            split_plan.execute()
        except Exception as ex:
            split_plan.rollback()
            raise TrestleError(f'Could not perform operation: {ex}')

    def _split_model(self, model: OscalBaseModel, element_paths: List[ElementPath]) -> Plan:
        """Split the model at the provided element paths."""
        raise NotImplementedError()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle write to file actions class."""

import pathlib

from tests import test_utils

from trestle.core.err import TrestleError
from trestle.core.models.actions import FileContentType, WriteFileAction
from trestle.core.models.elements import Element


def test_write_file_action_yaml(tmp_yaml_file: pathlib.Path, sample_target):
    """Test write yaml action."""
    element = Element(sample_target)

    try:
        # string path should error
        wa = WriteFileAction(tmp_yaml_file.as_posix(), element, FileContentType.YAML)
    except TrestleError:
        pass

    try:
        # invalid file extension should error
        wa = WriteFileAction(tmp_yaml_file, element, FileContentType.JSON)
    except TrestleError:
        pass

    try:
        # writing to a non-existing file will error
        wa = WriteFileAction(tmp_yaml_file, element, FileContentType.YAML)
        wa.execute()
    except TrestleError:
        pass

    tmp_yaml_file.touch()
    wa = WriteFileAction(tmp_yaml_file, element, FileContentType.YAML)
    wa.execute()
    test_utils.verify_file_content(tmp_yaml_file, element.get())


def test_write_file_rollback(tmp_yaml_file: pathlib.Path, sample_target):
    """Test rollback."""
    element = Element(sample_target)
    tmp_yaml_file.touch()
    wa = WriteFileAction(tmp_yaml_file, element, FileContentType.YAML)
    wa.execute()
    test_utils.verify_file_content(tmp_yaml_file, element.get())

    # verify the file content is not empty
    with open(tmp_yaml_file, 'r', encoding='utf8') as read_file:
        assert read_file.tell() >= 0

    wa.rollback()

    # verify the file content is empty
    with open(tmp_yaml_file, 'r', encoding='utf8') as read_file:
        assert read_file.tell() == 0


def test_write_existing_file_rollback(tmp_yaml_file, sample_target):
    """Test rollback."""
    # add some content
    tmp_yaml_file.touch()
    current_pos = 0
    with open(tmp_yaml_file, 'a+') as fp:
        fp.write('....\n')
        current_pos = fp.tell()

    # write to the file
    element = Element(sample_target)
    wa = WriteFileAction(tmp_yaml_file, element, FileContentType.YAML)
    wa.execute()

    # verify the file content is not empty
    with open(tmp_yaml_file, 'a+', encoding='utf8') as fp:
        assert fp.tell() > current_pos

    # rollback to the original
    wa.rollback()

    # # verify the file content is empty
    with open(tmp_yaml_file, 'a+', encoding='utf8') as fp:
        assert fp.tell() == current_pos



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle remove action class."""

from typing import List

from trestle.core.models.actions import RemoveAction, UpdateAction
from trestle.core.models.elements import Element, ElementPath
from trestle.oscal import target


def prepare_element(sample_target):
    """Prepare a target element for remove tests."""
    element = Element(sample_target)

    parties: List[target.Party] = []
    parties.append(
        target.Party(**{
            'uuid': 'ff47836c-877c-4007-bbf3-c9d9bd805000', 'party-name': 'TEST1', 'type': 'organization'
        })
    )
    parties.append(
        target.Party(**{
            'uuid': 'ee88836c-877c-4007-bbf3-c9d9bd805000', 'party-name': 'TEST2', 'type': 'organization'
        })
    )

    sub_element_path = ElementPath('metadata.parties.*')
    ac = UpdateAction(parties, element, sub_element_path)
    ac.execute()

    assert element.get_at(sub_element_path) == parties

    return element


def test_remove_action(sample_target):
    """Test remove action."""
    element = prepare_element(sample_target)
    sub_element_path = ElementPath('metadata.parties')
    prev_sub_element = element.get_at(sub_element_path)

    rac = RemoveAction(element, sub_element_path)

    rac.execute()

    assert element.get_at(sub_element_path) is None

    rac.rollback()

    assert element.get_at(sub_element_path) == prev_sub_element



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Common file system utilities."""

import json
import logging
import os
import pathlib

from trestle.core import const
from trestle.core.err import TrestleError

import yaml

logger = logging.getLogger(__name__)


def should_ignore(name) -> bool:
    """Check if the file or directory should be ignored or not."""
    return name[0] == '.' or name[0] == '_'


def ensure_directory(path):
    """
    Ensure the directory ```path``` exists.

    It creates the directories along the path if they do not exist.
    Arguments:
        path(str): The path to the directory

    Raises:
        AssertionError: If the directory path exists but is not a directory

    Returns: None
    """
    path = os.path.abspath(path)

    if not os.path.exists(path):
        os.makedirs(path)
    elif not os.path.isdir(path):
        raise AssertionError(f'Path `{path}` exists but is not a directory')


def is_valid_project_root(project_root: pathlib.Path) -> bool:
    """Check if the project root is a valid trestle project root."""
    if project_root is None or project_root == '' or len(project_root.parts) <= 0:
        return False

    trestle_dir = pathlib.Path.joinpath(project_root, const.TRESTLE_CONFIG_DIR)
    if trestle_dir.exists() and trestle_dir.is_dir():
        return True

    return False


def get_trestle_project_root(path: pathlib.Path) -> pathlib.Path:
    """Get the trestle project root folder in the path."""
    if path is None or len(path.parts) <= 0:
        return None

    current = path
    while len(current.parts) > 0:
        if is_valid_project_root(current):
            return current
        current = current.parent

    return None


def has_parent_path(sub_path: pathlib.Path, parent_path: pathlib.Path) -> bool:
    """Check if sub_path has the specified parent_dir path."""
    if parent_path is None or len(parent_path.parts) <= 0:
        return False

    # sub_path should be longer than parent path
    if len(sub_path.parts) < len(parent_path.parts):
        return False

    matched = True
    for i, part in enumerate(parent_path.parts):
        if part is not sub_path.parts[i]:
            matched = False
            break

    return matched


def has_trestle_project_in_path(path: pathlib.Path) -> bool:
    """Check if path has a valid trestle project among the parents."""
    trestle_project_root = get_trestle_project_root(path)
    return trestle_project_root is not None


def clean_project_sub_path(sub_path: pathlib.Path):
    """Clean all directories and files in the project sub sub.

    It ensures the sub_path is a child path in the project root.
    """
    if sub_path.exists():
        project_root = sub_path.parent
        if not has_trestle_project_in_path(project_root):
            raise TrestleError('Path to be cleaned is not a under valid Trestle project root')

        # clean all files/directories under sub_path
        if sub_path.is_dir():
            for item in pathlib.Path.iterdir(sub_path):
                if item.is_file():
                    item.unlink()
                elif item.is_dir():
                    clean_project_sub_path(item)
            sub_path.rmdir()
        # delete the sub_path
        elif sub_path.is_file():
            sub_path.unlink()


def load_file(file_name: str):
    """Load JSON or YAML file content."""
    _, file_extension = os.path.splitext(file_name)

    with open(file_name) as f:
        if file_extension == '.yaml':
            return yaml.load(f, yaml.FullLoader)
        elif file_extension == '.json':
            return json.load(f)
        else:
            raise TrestleError(f'Invalid file extension "{file_extension}"')


def find_node(data: dict, key: str, depth: int = 0, max_depth: int = 1, instance_type: type = list):
    """Find a node of an instance_type in the data recursively."""
    if depth > max_depth:
        return

    if key in data and isinstance(data[key], instance_type):
        yield data[key]

    for _, v in data.items():
        if isinstance(v, dict):
            for i in find_node(v, key, depth + 1, max_depth, instance_type):
                yield i
        elif isinstance(v, list):
            for item in v:
                for i in find_node(item, key, depth + 1, max_depth, instance_type):
                    yield i



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Starting point for the Trestle CLI."""

from ilcli import Command

from trestle.__init__ import __version__
from trestle.core.commands.add import AddCmd
from trestle.core.commands.assemble import AssembleCmd
from trestle.core.commands.create import CreateCmd
from trestle.core.commands.import_ import ImportCmd
from trestle.core.commands.init import InitCmd
from trestle.core.commands.merge import MergeCmd
from trestle.core.commands.remove import RemoveCmd
from trestle.core.commands.replicate import ReplicateCmd
from trestle.core.commands.split import SplitCmd
from trestle.core.commands.validate import ValidateCmd


class Trestle(Command):
    """Manage OSCAL files in a human friendly manner."""

    subcommands = [
        InitCmd, CreateCmd, SplitCmd, MergeCmd, ReplicateCmd, AddCmd, RemoveCmd, ValidateCmd, ImportCmd, AssembleCmd
    ]

    def _init_arguments(self):
        self.add_argument(
            '-V',
            '--version',
            help='Display the version of trestle.',
            action='version',
            version=f'Trestle version v{__version__}'
        )
        self.add_argument('-v', '--verbose', help='Display verbose output.', action='count', default=1)


def run():
    """Run the test cli."""
    exit(Trestle().run())


if __name__ == '__main__':
    run()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for models util module."""

import pathlib

import trestle.core.utils as mutils
import trestle.oscal.catalog as catalog
import trestle.oscal.target as ostarget

import yaml


def load_good_catalog():
    """Load nist 800-53 as a catalog example."""
    good_sample_path = pathlib.Path('nist-source/content/nist.gov/SP800-53/rev4/json/NIST_SP-800-53_rev4_catalog.json')

    assert (good_sample_path.exists())
    return catalog.Catalog.oscal_read(good_sample_path)


def test_get_elements():
    """Test getting flat list of elements."""
    good_sample = load_good_catalog()

    mdlist = mutils.get_elements_of_model_type(good_sample, catalog.Metadata)
    assert (type(mdlist) == list)
    # can only be 1 metadata
    assert (len(mdlist) == 1)
    assert (type(mdlist[0]) == catalog.Metadata)

    control_list = mutils.get_elements_of_model_type(good_sample, catalog.Control)
    assert (len(control_list) >= 1)
    group_list = mutils.get_elements_of_model_type(good_sample, catalog.Group)
    assert (len(group_list) >= 2)


def test_has_no_duplicate_values_generic():
    """Test presence of duplicate uuid."""
    # test with pydantic catalog
    cat = load_good_catalog()
    assert mutils.has_no_duplicate_values_generic(cat, 'uuid')

    yaml_path = pathlib.Path('tests/data/yaml')

    # test with valid pydantic target
    good_target_path = yaml_path / 'good_target.yaml'
    good_target = ostarget.TargetDefinition.oscal_read(good_target_path)
    loe = mutils.find_values_by_name(good_target, 'uuid')
    assert len(loe) == 5
    assert mutils.has_no_duplicate_values_by_name(good_target, 'uuid')

    # test with pydantic target containing duplicates
    bad_target_path = yaml_path / 'bad_target_dup_uuid.yaml'
    bad_target = ostarget.TargetDefinition.oscal_read(bad_target_path)
    assert not mutils.has_no_duplicate_values_by_name(bad_target, 'uuid')

    # test duplicates with raw yaml target, non-pydantic
    read_file = bad_target_path.open('r', encoding='utf8')
    bad_target_yaml = yaml.load(read_file, Loader=yaml.Loader)
    assert not mutils.has_no_duplicate_values_generic(bad_target_yaml, 'uuid')


def test_has_no_duplicate_values_pydantic():
    """Test presence of duplicate values in pydantic objects."""
    # test with pydantic catalog - only one instance of Metadata
    cat = load_good_catalog()
    assert mutils.has_no_duplicate_values_by_type(cat, catalog.Metadata)

    yaml_path = pathlib.Path('tests/data/yaml')

    # test presence of many duplicate properties
    good_target_path = yaml_path / 'good_target.yaml'
    good_target = ostarget.TargetDefinition.oscal_read(good_target_path)
    assert not mutils.has_no_duplicate_values_by_type(good_target, ostarget.Prop)



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for models module."""

import pathlib
from datetime import datetime
from uuid import uuid4

from trestle.core import const
from trestle.core import parser
from trestle.oscal import catalog
from trestle.oscal import target

import yaml

yaml_path = pathlib.Path('tests/data/yaml/')
encoding = 'utf8'


def test_parse_dict():
    """Test parse_dict."""
    file_name = 'good_target.yaml'

    with open(pathlib.Path.joinpath(yaml_path, file_name), 'r', encoding=encoding) as f:
        data = yaml.load(f, yaml.FullLoader)
        target = parser.parse_dict(data['target-definition'], 'trestle.oscal.target.TargetDefinition')
        assert target is not None


def test_to_class_name():
    """Test to_class_name."""
    tests = [
        {
            'name': 'catalog',
            'expected': 'Catalog',
        },
        {
            'name': 'target-definition',
            'expected': 'TargetDefinition',
        },
    ]

    for test in tests:
        assert parser.to_class_name(test['name']) == test['expected']


def test_to_full_model_name():
    """Test to_full_model_name."""
    tests = [
        {
            'root_key': 'catalog',
            'name': None,
            'expected': f'{const.PACKAGE_OSCAL}.catalog.Catalog',
        }, {
            'root_key': 'catalog',
            'name': 'group',
            'expected': f'{const.PACKAGE_OSCAL}.catalog.Group',
        }, {
            'root_key': 'target-definition',
            'name': None,
            'expected': f'{const.PACKAGE_OSCAL}.target.TargetDefinition',
        },
        {
            'root_key': 'target',
            'name': 'target-definition',
            'expected': f'{const.PACKAGE_OSCAL}.target.TargetDefinition',
        }, {
            'root_key': 'invalid',
            'name': None,
            'expected': None,
        }
    ]

    for test in tests:
        model_name = parser.to_full_model_name(test['root_key'], test['name'])
        assert model_name == test['expected']


def test_parse_file():
    """Test parse_file."""
    file_name = 'good_target.yaml'

    tests = [
        {
            'model_name': f'{const.PACKAGE_OSCAL}.target.TargetDefinition', 'expected': 'TargetDefinition'
        }, {
            'model_name': None, 'expected': 'TargetDefinition'
        }
    ]

    for test in tests:
        target = parser.parse_file(pathlib.Path.joinpath(yaml_path, file_name), model_name=test['model_name'])

        assert type(target).__name__ == test['expected']


def test_class_to_oscal_json():
    """Pydantic makes classes in PascalCase. All Oscal names are in lowercase-hyphenated."""
    class_name_1 = 'Catalog'
    oscal_name_1 = 'catalog'
    class_name_2 = 'ComponentDefinition'
    oscal_name_2 = 'component-definition'
    class_name_3 = 'SecurityImpactLevel'
    oscal_name_3 = 'security-impact-level'

    assert (oscal_name_1 == parser.class_to_oscal(class_name_1, 'json'))
    assert (oscal_name_2 == parser.class_to_oscal(class_name_2, 'json'))
    assert (oscal_name_3 == parser.class_to_oscal(class_name_3, 'json'))


def test_class_to_oscal_field():
    """Pydantic makes classes in PascalCase. All Oscal names are in lowercase-hyphenated."""
    class_name_1 = 'Catalog'
    oscal_name_1 = 'catalog'
    class_name_2 = 'ComponentDefinition'
    oscal_name_2 = 'component_definition'
    class_name_3 = 'SecurityImpactLevel'
    oscal_name_3 = 'security_impact_level'

    assert (oscal_name_1 == parser.class_to_oscal(class_name_1, 'field'))
    assert (oscal_name_2 == parser.class_to_oscal(class_name_2, 'field'))
    assert (oscal_name_3 == parser.class_to_oscal(class_name_3, 'field'))


def test_wrap_for_output():
    """Test that an output object is wrapped and contains the appropriate content."""
    m = catalog.Metadata(
        **{
            'title': 'my cool catalog', 'last-modified': datetime.now(), 'version': '0.0.1', 'oscal-version': '1.0.0'
        }
    )

    c = catalog.Catalog(metadata=m, uuid=str(uuid4()))

    wrapped = parser.wrap_for_output(c)
    assert (wrapped.catalog.metadata.title == c.metadata.title)

    # Test with target definition
    t_m = target.Metadata(
        **{
            'title': 'my cool target definition',
            'last-modified': datetime.now(),
            'version': '0.0.1',
            'oscal-version': '1.0.0'
        }
    )

    t = target.TargetDefinition(metadata=t_m)
    wrapped = parser.wrap_for_output(t)
    assert (wrapped.target_definition.metadata.title == t.metadata.title)


def test_wrap_for_input():
    """Test input for object that has a hyphen it it's name."""
    json_path = pathlib.Path('tests/data/json/sample-target-definition.json')
    parser.wrap_for_input(target.TargetDefinition).parse_file(json_path).target_definition


def test_pascal_case_split():
    """Test whether PascalCase objects are getting split correctly."""
    one = 'One'
    two = 'TwoTwo'
    three = 'ThreeThreeThree'
    assert (len(parser.pascal_case_split(one)) == 1)
    assert (len(parser.pascal_case_split(two)) == 2)
    assert (len(parser.pascal_case_split(three)) == 3)
    # TODO: Add negative test cases.
